from collections import defaultdict
import logging
from typing import Dict, List, Tuple, Set, Optional, Iterable, Union, Type, TYPE_CHECKING

import networkx

import ailment

from ...knowledge_base import KnowledgeBase
from ...codenode import BlockNode
from ...utils import timethis
from ...calling_conventions import SimRegArg, SimStackArg, SimFunctionArgument
from ...sim_type import SimTypeChar, SimTypeInt, SimTypeLongLong, SimTypeShort, SimTypeFunction, SimTypeBottom
from ...sim_variable import SimVariable, SimStackVariable, SimRegisterVariable, SimMemoryVariable
from ...knowledge_plugins.key_definitions.constants import OP_BEFORE
from ...procedures.stubs.UnresolvableCallTarget import UnresolvableCallTarget
from ...procedures.stubs.UnresolvableJumpTarget import UnresolvableJumpTarget
from .. import Analysis, register_analysis
from ..cfg.cfg_base import CFGBase
from .ailgraph_walker import AILGraphWalker
from .ailblock_walker import AILBlockWalker
from .optimization_passes import get_default_optimization_passes, OptimizationPassStage

if TYPE_CHECKING:
    from angr.knowledge_plugins.cfg import CFGModel
    from .decompilation_cache import DecompilationCache
    from .peephole_optimizations import PeepholeOptimizationStmtBase, PeepholeOptimizationExprBase


l = logging.getLogger(name=__name__)


class Clinic(Analysis):
    """
    A Clinic deals with AILments.
    """
    def __init__(self, func,
                 remove_dead_memdefs=False,
                 exception_edges=False,
                 sp_tracker_track_memory=True,
                 optimization_passes=None,
                 cfg=None,
                 peephole_optimizations: Optional[Iterable[Union[Type['PeepholeOptimizationStmtBase'],Type['PeepholeOptimizationExprBase']]]]=None,  # pylint:disable=line-too-long
                 must_struct: Optional[Set[str]]=None,
                 variable_kb=None,
                 reset_variable_names=False,
                 cache: Optional['DecompilationCache']=None,
                 ):
        if not func.normalized:
            raise ValueError("Decompilation must work on normalized function graphs.")

        self.function = func

        self.graph = None
        self.arg_list = None
        self.variable_kb = variable_kb

        self._func_graph: Optional[networkx.DiGraph] = None
        self._ail_manager = None
        self._blocks_by_addr_and_size = { }

        self._remove_dead_memdefs = remove_dead_memdefs
        self._exception_edges = exception_edges
        self._sp_tracker_track_memory = sp_tracker_track_memory
        self._cfg: Optional['CFGModel'] = cfg
        self.peephole_optimizations = peephole_optimizations
        self._must_struct = must_struct
        self._reset_variable_names = reset_variable_names
        self._cache = cache

        # sanity checks
        if not self.kb.functions:
            l.warning('No function is available in kb.functions. It will lead to a suboptimal conversion result.')

        if optimization_passes is not None:
            self._optimization_passes = optimization_passes
        else:
            self._optimization_passes = get_default_optimization_passes(self.project.arch, self.project.simos.name)
            l.debug("Get %d optimization passes for the current binary.", len(self._optimization_passes))

        self._analyze()

    #
    # Public methods
    #

    def block(self, addr, size):
        """
        Get the converted block at the given specific address with the given size.

        :param int addr:
        :param int size:
        :return:
        """

        try:
            return self._blocks_by_addr_and_size[(addr, size)]
        except KeyError:
            return None

    def dbg_repr(self):
        """

        :return:
        """

        s = ""

        for block in sorted(self.graph.nodes(), key=lambda x: x.addr):
            s += str(block) + "\n\n"

        return s

    #
    # Private methods
    #

    def _analyze(self):

        # Set up the function graph according to configurations
        self._update_progress(0., text="Setting up function graph")
        self._set_function_graph()

        # Remove alignment blocks
        self._update_progress(5., text="Removing alignment blocks")
        self._remove_alignment_blocks()

        # if the graph is empty, don't continue
        if not self._func_graph:
            return

        # Make sure calling conventions of all functions have been recovered
        self._update_progress(10., text="Recovering calling conventions")
        self._recover_calling_conventions()

        # initialize the AIL conversion manager
        self._ail_manager = ailment.Manager(arch=self.project.arch)

        # Track stack pointers
        self._update_progress(15., text="Tracking stack pointers")
        spt = self._track_stack_pointers()

        # Convert VEX blocks to AIL blocks and then simplify them

        self._update_progress(20., text="Converting VEX to AIL")
        self._convert_all()

        ail_graph = self._make_ailgraph()

        # Fix "fake" indirect jumps and calls
        self._update_progress(25., text="Analyzing simple indirect jumps")
        ail_graph = self._replace_single_target_indirect_transitions(ail_graph)

        # Make returns
        self._update_progress(30., text="Making return sites")
        if self.function.prototype is None or not isinstance(self.function.prototype.returnty, SimTypeBottom):
            ail_graph = self._make_returns(ail_graph)

        # Simplify blocks
        # we never remove dead memory definitions before making callsites. otherwise stack arguments may go missing
        # before they are recognized as stack arguments.
        self._update_progress(35., text="Simplifying blocks 1")
        ail_graph = self._simplify_blocks(ail_graph, stack_pointer_tracker=spt, remove_dead_memdefs=False)

        # Run simplification passes
        self._update_progress(40., text="Running simplifications 1")
        ail_graph = self._run_simplification_passes(ail_graph,
                                                    stage=OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION)

        # Simplify the entire function for the first time
        self._update_progress(45., text="Simplifying function 1")
        self._simplify_function(ail_graph, remove_dead_memdefs=False, unify_variables=False)

        # clear _blocks_by_addr_and_size so no one can use it again
        # TODO: Totally remove this dict
        self._blocks_by_addr_and_size = None

        # Make call-sites
        self._update_progress(50., text="Making callsites")
        _, stackarg_offsets = self._make_callsites(ail_graph, stack_pointer_tracker=spt)

        # Simplify the entire function for the second time
        self._update_progress(55., text="Simplifying function 2")
        self._simplify_function(ail_graph, remove_dead_memdefs=self._remove_dead_memdefs,
                                stack_arg_offsets=stackarg_offsets, unify_variables=True)

        # After global optimization, there might be more chances for peephole optimizations.
        # Simplify blocks for the second time
        self._update_progress(60., text="Simplifying blocks 2")
        ail_graph = self._simplify_blocks(ail_graph, remove_dead_memdefs=self._remove_dead_memdefs,
                                          stack_pointer_tracker=spt)

        # Simplify the entire function for the third time
        self._update_progress(65., text="Simplifying function 3")
        self._simplify_function(ail_graph, remove_dead_memdefs=self._remove_dead_memdefs,
                                stack_arg_offsets=stackarg_offsets, unify_variables=True)

        # Make function arguments
        self._update_progress(70., text="Making argument list")
        arg_list = self._make_argument_list()

        # Run simplification passes
        self._update_progress(75., text="Running simplifications 2")
        ail_graph = self._run_simplification_passes(ail_graph, stage=OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION)

        # Recover variables on AIL blocks
        self._update_progress(80., text="Recovering variables")
        variable_kb = self._recover_and_link_variables(ail_graph, arg_list)

        # Make function prototype
        self._update_progress(85., text="Making function prototype")
        self._make_function_prototype(arg_list, variable_kb)

        # Run simplification passes
        self._update_progress(90., text="Running simplifications 3")
        ail_graph = self._run_simplification_passes(ail_graph, stage=OptimizationPassStage.AFTER_VARIABLE_RECOVERY)

        self.graph = ail_graph
        self.arg_list = arg_list
        self.variable_kb = variable_kb

    @timethis
    def _set_function_graph(self):
        self._func_graph = self.function.graph_ex(exception_edges=self._exception_edges)

    @timethis
    def _remove_alignment_blocks(self):
        """
        Alignment blocks are basic blocks that only consist of nops. They should not be included in the graph.
        """
        for node in list(self._func_graph.nodes()):
            if self._func_graph.in_degree(node) == 0 and \
                    CFGBase._is_noop_block(self.project.arch, self.project.factory.block(node.addr, node.size)):
                self._func_graph.remove_node(node)

    @timethis
    def _recover_calling_conventions(self):
        self.project.analyses.CompleteCallingConventions()

    @timethis
    def _track_stack_pointers(self):
        """
        For each instruction, track its stack pointer offset and stack base pointer offset.

        :return: None
        """

        regs = {self.project.arch.sp_offset}
        if hasattr(self.project.arch, 'bp_offset') and self.project.arch.bp_offset is not None:
            regs.add(self.project.arch.bp_offset)
        spt = self.project.analyses.StackPointerTracker(self.function, regs, track_memory=self._sp_tracker_track_memory)
        if spt.inconsistent_for(self.project.arch.sp_offset):
            l.warning("Inconsistency found during stack pointer tracking. Decompilation results might be incorrect.")
        return spt

    @timethis
    def _convert_all(self):
        """
        Convert all VEX blocks in the function graph to AIL blocks, and fill self._blocks.

        :return:    None
        """

        for block_node in self._func_graph.nodes():
            ail_block = self._convert(block_node)

            if type(ail_block) is ailment.Block:
                self._blocks_by_addr_and_size[(block_node.addr, block_node.size)] = ail_block

    def _convert(self, block_node):
        """
        Convert a VEX block to an AIL block.

        :param block_node:  A BlockNode instance.
        :return:            An converted AIL block.
        :rtype:             ailment.Block
        """

        if not type(block_node) is BlockNode:
            return block_node

        block = self.project.factory.block(block_node.addr, block_node.size)

        ail_block = ailment.IRSBConverter.convert(block.vex, self._ail_manager)
        return ail_block

    @timethis
    def _replace_single_target_indirect_transitions(self, ail_graph: networkx.DiGraph) -> networkx.DiGraph:
        """
        Remove single-target indirect jumps and calls and replace them with direct jumps or calls.
        """
        if self._cfg is None:
            return ail_graph

        for block in ail_graph.nodes():
            if not block.statements:
                continue
            last_stmt = block.statements[-1]
            if isinstance(last_stmt, ailment.Stmt.Call) and not isinstance(last_stmt.target, ailment.Expr.Const):
                # indirect call
                # consult CFG to see if this is a call with a single successor
                node = self._cfg.get_any_node(block.addr)
                if node is None:
                    continue
                successors = self._cfg.get_successors(node, excluding_fakeret=True, jumpkind='Ijk_Call')
                if len(successors) == 1 and \
                        not isinstance(self.project.hooked_by(successors[0].addr), UnresolvableCallTarget):
                    # found a single successor - replace the last statement
                    new_last_stmt = last_stmt.copy()
                    new_last_stmt.target = ailment.Expr.Const(None, None, successors[0].addr, last_stmt.target.bits)
                    block.statements[-1] = new_last_stmt

            elif isinstance(last_stmt, ailment.Stmt.Jump) and not isinstance(last_stmt.target, ailment.Expr.Const):
                # indirect jump
                # consult CFG to see if there is a jump with a single successor
                node = self._cfg.get_any_node(block.addr)
                if node is None:
                    continue
                successors = self._cfg.get_successors(node, excluding_fakeret=True, jumpkind='Ijk_Boring')
                if len(successors) == 1 and \
                        not isinstance(self.project.hooked_by(successors[0].addr), UnresolvableJumpTarget):
                    # found a single successor - replace the last statement
                    new_last_stmt = last_stmt.copy()
                    new_last_stmt.target = ailment.Expr.Const(None, None, successors[0].addr, last_stmt.target.bits)
                    block.statements[-1] = new_last_stmt

        return ail_graph

    @timethis
    def _make_ailgraph(self) -> networkx.DiGraph:
        graph = self._function_graph_to_ail_graph(self._func_graph)
        return graph

    @timethis
    def _simplify_blocks(self, ail_graph: networkx.DiGraph, remove_dead_memdefs=False, stack_pointer_tracker=None):
        """
        Simplify all blocks in self._blocks.

        :param ail_graph:               The AIL function graph.
        :param stack_pointer_tracker:   The RegisterDeltaTracker analysis instance.
        :return:                        None
        """

        blocks_by_addr_and_idx: Dict[Tuple[int,Optional[int]],ailment.Block] = { }

        for ail_block in ail_graph.nodes():
            simplified = self._simplify_block(ail_block, remove_dead_memdefs=remove_dead_memdefs,
                                              stack_pointer_tracker=stack_pointer_tracker)
            key = ail_block.addr, ail_block.idx
            blocks_by_addr_and_idx[key] = simplified

        # update blocks_map to allow node_addr to node lookup
        def _replace_node_handler(node):
            key = node.addr, node.idx
            if key in blocks_by_addr_and_idx:
                return blocks_by_addr_and_idx[key]
            return None

        AILGraphWalker(ail_graph, _replace_node_handler, replace_nodes=True).walk()

        return ail_graph

    def _simplify_block(self, ail_block, remove_dead_memdefs=False, stack_pointer_tracker=None):
        """
        Simplify a single AIL block.

        :param ailment.Block ail_block: The AIL block to simplify.
        :param stack_pointer_tracker:   The RegisterDeltaTracker analysis instance.
        :return:                        A simplified AIL block.
        """

        simp = self.project.analyses.AILBlockSimplifier(
            ail_block,
            remove_dead_memdefs=remove_dead_memdefs,
            stack_pointer_tracker=stack_pointer_tracker,
            peephole_optimizations=self.peephole_optimizations,
        )
        return simp.result_block

    @timethis
    def _simplify_function(self, ail_graph, remove_dead_memdefs=False, stack_arg_offsets=None, unify_variables=False,
                           max_iterations: int=8) -> None:
        """
        Simplify the entire function until it reaches a fixed point.
        """

        for _ in range(max_iterations):
            simplified = self._simplify_function_once(ail_graph, remove_dead_memdefs=remove_dead_memdefs,
                                                      unify_variables=unify_variables,
                                                      stack_arg_offsets=stack_arg_offsets)
            if not simplified:
                break

    @timethis
    def _simplify_function_once(self, ail_graph, remove_dead_memdefs=False, stack_arg_offsets=None,
                                unify_variables=False):
        """
        Simplify the entire function once.

        :return:    None
        """

        simp = self.project.analyses.AILSimplifier(
            self.function,
            func_graph=ail_graph,
            remove_dead_memdefs=remove_dead_memdefs,
            unify_variables=unify_variables,
            stack_arg_offsets=stack_arg_offsets,
        )
        # the function graph has been updated at this point
        return simp.simplified

    @timethis
    def _run_simplification_passes(self, ail_graph, stage:int=OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION):

        addr_and_idx_to_blocks: Dict[Tuple[int,Optional[int]],ailment.Block] = { }
        addr_to_blocks: Dict[int,Set[ailment.Block]] = defaultdict(set)

        # update blocks_map to allow node_addr to node lookup
        def _updatedict_handler(node):
            addr_and_idx_to_blocks[(node.addr, node.idx)] = node
            addr_to_blocks[node.addr].add(node)
        AILGraphWalker(ail_graph, _updatedict_handler).walk()

        # Run each pass
        for pass_ in self._optimization_passes:

            if pass_.STAGE != stage:
                continue

            analysis = getattr(self.project.analyses, pass_.__name__)

            a = analysis(self.function, blocks_by_addr=addr_to_blocks, blocks_by_addr_and_idx=addr_and_idx_to_blocks,
                         graph=ail_graph)
            if a.out_graph:
                # use the new graph
                ail_graph = a.out_graph

        return ail_graph

    @timethis
    def _make_argument_list(self) -> List[SimVariable]:
        if self.function.calling_convention is not None and self.function.prototype is not None:
            args: List[SimFunctionArgument] = self.function.calling_convention.arg_locs(self.function.prototype)
            arg_vars: List[SimVariable] = [ ]
            if args:
                for idx, arg in enumerate(args):
                    if isinstance(arg, SimRegArg):
                        argvar = SimRegisterVariable(
                            self.project.arch.registers[arg.reg_name][0],
                            arg.size,
                            ident="arg_%d" % idx,
                            name="a%d" % idx,
                            region=self.function.addr,
                        )
                    elif isinstance(arg, SimStackArg):
                        argvar = SimStackVariable(
                            arg.stack_offset,
                            arg.size,
                            base='bp',
                            ident="arg_%d" % idx,
                            name="a%d" % idx,
                            region=self.function.addr,
                        )
                    else:
                        raise TypeError("Unsupported function argument type %s." % type(arg))
                    arg_vars.append(argvar)
            return arg_vars
        return [ ]

    @timethis
    def _make_callsites(self, ail_graph, stack_pointer_tracker=None):
        """
        Simplify all function call statements.

        :return:    None
        """

        # Computing reaching definitions
        rd = self.project.analyses.ReachingDefinitions(subject=self.function, func_graph=ail_graph,
                                                       observe_callback=self._make_callsites_rd_observe_callback)

        class TempClass:  # pylint:disable=missing-class-docstring
            stack_arg_offsets = set()

        def _handler(block):
            csm = self.project.analyses.AILCallSiteMaker(block,
                                                         reaching_definitions=rd,
                                                         stack_pointer_tracker=stack_pointer_tracker,
                                                         ail_manager=self._ail_manager,
                                                         )
            if csm.stack_arg_offsets is not None:
                TempClass.stack_arg_offsets |= csm.stack_arg_offsets
            if csm.result_block:
                if csm.result_block != block:
                    ail_block = csm.result_block
                    simp = self.project.analyses.AILBlockSimplifier(ail_block,
                                                                    stack_pointer_tracker=stack_pointer_tracker,
                                                                    peephole_optimizations=self.peephole_optimizations,
                                                                    stack_arg_offsets=csm.stack_arg_offsets,
                                                                    )
                    return simp.result_block
            return None

        AILGraphWalker(ail_graph, _handler, replace_nodes=True).walk()

        return ail_graph, TempClass.stack_arg_offsets

    @timethis
    def _make_returns(self, ail_graph: networkx.DiGraph) -> networkx.DiGraph:
        """
        Work on each return statement and fill in its return expressions.
        """

        if self.function.calling_convention is None:
            # unknown calling convention. cannot do much about return expressions.
            return ail_graph

        # Block walker

        def _handle_Return(stmt_idx: int, stmt: ailment.Stmt.Return, block: Optional[ailment.Block]):  # pylint:disable=unused-argument
            if block is not None \
                    and not stmt.ret_exprs \
                    and self.function.prototype is not None \
                    and type(self.function.prototype.returnty) is not SimTypeBottom:
                new_stmt = stmt.copy()
                ret_val = self.function.calling_convention.return_val(self.function.prototype.returnty)
                if isinstance(ret_val, SimRegArg):
                    reg = self.project.arch.registers[ret_val.reg_name]
                    new_stmt.ret_exprs.append(ailment.Expr.Register(
                        self._next_atom(),
                        None,
                        reg[0],
                        reg[1] * self.project.arch.byte_width,
                        reg_name=self.project.arch.translate_register_name(reg[0], reg[1])
                    ))
                else:
                    l.warning("Unsupported type of return expression %s.", type(ret_val))
                block.statements[stmt_idx] = new_stmt


        def _handler(block):
            walker = AILBlockWalker()
            # we don't need to handle any statement besides Returns
            walker.stmt_handlers.clear()
            walker.expr_handlers.clear()
            walker.stmt_handlers[ailment.Stmt.Return] = _handle_Return
            walker.walk(block)

        # Graph walker

        AILGraphWalker(ail_graph, _handler, replace_nodes=True).walk()

        return ail_graph

    @timethis
    def _make_function_prototype(self, arg_list: List[SimVariable], variable_kb):
        if self.function.prototype is not None and not self.function.is_prototype_guessed:
            # do not overwrite an existing function prototype
            # if you want to re-generate the prototype, clear the existing one first
            return

        variables = variable_kb.variables[self.function.addr]
        func_args = [ ]
        for arg in arg_list:
            func_arg = None
            arg_ty = variables.get_variable_type(arg)
            if arg_ty is None:
                # determine type based on size
                if isinstance(arg, (SimRegisterVariable, SimStackVariable)):
                    if arg.size == 1:
                        func_arg = SimTypeChar()
                    elif arg.size == 2:
                        func_arg = SimTypeShort()
                    elif arg.size == 4:
                        func_arg = SimTypeInt()
                    elif arg.size == 8:
                        func_arg = SimTypeLongLong()
                    else:
                        l.warning("Unsupported argument size %d.", arg.size)
            else:
                func_arg = arg_ty

            func_args.append(func_arg)

        # TODO: need a new method of determining whether a function returns void
        returnty = SimTypeInt()

        self.function.prototype = SimTypeFunction(func_args, returnty).with_arch(self.project.arch)
        self.function.is_prototype_guessed = False

    @timethis
    def _recover_and_link_variables(self, ail_graph, arg_list):

        # variable recovery
        tmp_kb = KnowledgeBase(self.project) if self.variable_kb is None else self.variable_kb
        vr = self.project.analyses.VariableRecoveryFast(self.function,  # pylint:disable=unused-variable
                                                        func_graph=ail_graph, kb=tmp_kb, track_sp=False,
                                                        func_args=arg_list)
        # get ground-truth types
        var_manager = tmp_kb.variables[self.function.addr]
        groundtruth = {}
        for variable in var_manager.variables_with_manual_types:
            vartype = var_manager.types.get(variable, None)
            if vartype is not None:
                groundtruth[vr.var_to_typevar[variable]] = vartype
        # clean up existing types for this function
        var_manager.remove_types()
        # TODO: Type inference for global variables
        # run type inference
        if self._must_struct:
            must_struct = set()
            for var, typevar in vr.var_to_typevar.items():
                if var.ident in self._must_struct:
                    must_struct.add(typevar)
        else:
            must_struct = None
        try:
            tp = self.project.analyses.Typehoon(vr.type_constraints, kb=tmp_kb, var_mapping=vr.var_to_typevar,
                                                must_struct=must_struct, ground_truth=groundtruth)
            # tp.pp_constraints()
            # tp.pp_solution()
            tp.update_variable_types(self.function.addr, vr.var_to_typevar)
            tp.update_variable_types('global', vr.var_to_typevar)
        except Exception:  # pylint:disable=broad-except
            l.warning("Typehoon analysis failed. Variables will not have types. Please report to GitHub.",
                      exc_info=True)

        # Unify SSA variables
        tmp_kb.variables.global_manager.assign_variable_names(labels=self.kb.labels, types={SimMemoryVariable})
        var_manager.unify_variables()
        var_manager.assign_unified_variable_names(
            labels=self.kb.labels,
            reset=self._reset_variable_names,
        )

        # Link variables to each statement
        for block in ail_graph.nodes():
            self._link_variables_on_block(block, tmp_kb)

        if self._cache is not None:
            self._cache.type_constraints = vr.type_constraints
            self._cache.var_to_typevar = vr.var_to_typevar

        return tmp_kb

    def _link_variables_on_block(self, block, kb):
        """
        Link atoms (AIL expressions) in the given block to corresponding variables identified previously.

        :param ailment.Block block: The AIL block to work on.
        :return:                    None
        """

        variable_manager = kb.variables[self.function.addr]
        global_variables = kb.variables['global']

        for stmt_idx, stmt in enumerate(block.statements):
            stmt_type = type(stmt)
            if stmt_type is ailment.Stmt.Store:
                # find a memory variable
                mem_vars = variable_manager.find_variables_by_atom(block.addr, stmt_idx, stmt)
                if len(mem_vars) == 1:
                    stmt.variable, stmt.offset = next(iter(mem_vars))
                else:
                    # check if the dest address is a variable
                    stmt: ailment.Stmt.Store
                    # special handling for constant addresses
                    if isinstance(stmt.addr, ailment.Expr.Const):
                        # global variable?
                        variables = global_variables.get_global_variables(stmt.addr.value)
                        if variables:
                            var = next(iter(variables))
                            stmt.variable = var
                            stmt.offset = 0
                    else:
                        self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                                     stmt.addr)
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.data)

            elif stmt_type is ailment.Stmt.Assignment:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.dst)
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.src)

            elif stmt_type is ailment.Stmt.ConditionalJump:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.condition)

            elif stmt_type is ailment.Stmt.Call:
                self._link_variables_on_call(variable_manager, global_variables, block, stmt_idx, stmt, is_expr=False)

            elif stmt_type is ailment.Stmt.Return:
                self._link_variables_on_return(variable_manager, global_variables, block, stmt_idx, stmt)

    def _link_variables_on_return(self, variable_manager, global_variables, block: ailment.Block, stmt_idx: int,
                                  stmt: ailment.Stmt.Return):
        if stmt.ret_exprs:
            for ret_expr in stmt.ret_exprs:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, ret_expr)

    def _link_variables_on_call(self, variable_manager, global_variables, block, stmt_idx, stmt, is_expr=False):
        if stmt.args:
            for arg in stmt.args:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                             arg)
        if not is_expr and stmt.ret_expr:
            self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, stmt.ret_expr)

    def _link_variables_on_expr(self, variable_manager, global_variables, block, stmt_idx, stmt, expr):
        """
        Link atoms (AIL expressions) in the given expression to corresponding variables identified previously.

        :param variable_manager:    Variable manager of the function.
        :param ailment.Block block: AIL block.
        :param int stmt_idx:        ID of the statement.
        :param stmt:                The AIL statement that `expr` belongs to.
        :param expr:                The AIl expression to work on.
        :return:                    None
        """

        if type(expr) is ailment.Expr.Register:
            # find a register variable
            reg_vars = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            final_reg_vars = set()
            if len(reg_vars) > 1:
                # take phi variables
                for reg_var in reg_vars:
                    if variable_manager.is_phi_variable(reg_var[0]):
                        final_reg_vars.add(reg_var)
            else:
                final_reg_vars = reg_vars
            if len(final_reg_vars) == 1:
                reg_var, offset = next(iter(final_reg_vars))
                expr.variable = reg_var
                expr.variable_offset = offset

        elif type(expr) is ailment.Expr.Load:
            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            if len(variables) == 0:
                # if it's a constant addr, maybe it's referencing an extern location
                if isinstance(expr.addr, ailment.Expr.Const):
                    symbol = self.project.loader.find_symbol(expr.addr.value)
                    if symbol is not None:
                        # Create a new global variable if there isn't one already
                        global_vars = global_variables.get_global_variables(symbol.rebased_addr)
                        if global_vars:
                            global_var = next(iter(global_vars))
                        else:
                            global_var = SimMemoryVariable(symbol.rebased_addr, symbol.size, name=symbol.name)
                            global_variables.add_variable('global', global_var.addr, global_var)
                        expr.variable = global_var
                        expr.variable_offset = 0
                else:
                    # this is a local variable
                    self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, expr.addr)
                    if 'reference_variable' in expr.addr.tags and expr.addr.reference_variable is not None:
                        # copy over the variable to this expr since the variable on a constant is supposed to be a
                        # reference variable.
                        expr.variable = expr.addr.reference_variable
                        expr.variable_offset = expr.addr.reference_variable_offset
            else:
                if len(variables) > 1:
                    l.error("More than one variable are available for atom %s. Consider fixing it using phi nodes.",
                            expr
                            )
                var, offset = next(iter(variables))
                expr.variable = var
                expr.variable_offset = offset

        elif type(expr) is ailment.Expr.BinaryOp:
            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            if len(variables) == 1:
                var, offset = next(iter(variables))
                expr.variable = var
                expr.variable_offset = offset
            else:
                self._link_variables_on_expr(
                    variable_manager, global_variables, block, stmt_idx, stmt, expr.operands[0])
                self._link_variables_on_expr(
                    variable_manager, global_variables, block, stmt_idx, stmt, expr.operands[1])

        elif type(expr) is ailment.Expr.UnaryOp:
            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            if len(variables) == 1:
                var, offset = next(iter(variables))
                expr.variable = var
                expr.variable_offset = offset
            else:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, expr.operands)

        elif type(expr) is ailment.Expr.Convert:
            self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt, expr.operand)

        elif type(expr) is ailment.Expr.ITE:
            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            if len(variables) == 1:
                var, offset = next(iter(variables))
                expr.variable = var
                expr.variable_offset = offset
            else:
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                             expr.cond)
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                             expr.iftrue)
                self._link_variables_on_expr(variable_manager, global_variables, block, stmt_idx, stmt,
                                             expr.iftrue)

        elif isinstance(expr, ailment.Expr.BasePointerOffset):
            variables = variable_manager.find_variables_by_atom(block.addr, stmt_idx, expr)
            if len(variables) == 1:
                var, offset = next(iter(variables))
                expr.variable = var
                expr.variable_offset = offset

        elif isinstance(expr, ailment.Expr.Const):
            # global variable?
            variables = global_variables.get_global_variables(expr.value)
            if variables:
                var = next(iter(variables))
                expr.tags['reference_variable'] = var
                expr.tags['reference_variable_offset'] = None

        elif isinstance(expr, ailment.Stmt.Call):
            self._link_variables_on_call(variable_manager, global_variables, block, stmt_idx, expr, is_expr=True)

    def _function_graph_to_ail_graph(self, func_graph, blocks_by_addr_and_size=None):

        if blocks_by_addr_and_size is None:
            blocks_by_addr_and_size = self._blocks_by_addr_and_size

        node_to_block_mapping = {}
        graph = networkx.DiGraph()

        for node in func_graph.nodes():
            ail_block = blocks_by_addr_and_size.get((node.addr, node.size), node)
            node_to_block_mapping[node] = ail_block

            if ail_block is not None:
                graph.add_node(ail_block)

        for src_node, dst_node, data in func_graph.edges(data=True):
            src = node_to_block_mapping[src_node]
            dst = node_to_block_mapping[dst_node]

            if dst is not None:
                graph.add_edge(src, dst, **data)

        return graph

    def _next_atom(self) -> int:
        return self._ail_manager.next_atom()

    @staticmethod
    def _make_callsites_rd_observe_callback(ob_type, **kwargs):
        if ob_type != 'insn':
            return False
        stmt = kwargs.pop('stmt')
        op_type = kwargs.pop('op_type')
        return isinstance(stmt, ailment.Stmt.Call) and op_type == OP_BEFORE


register_analysis(Clinic, 'Clinic')



import logging

from ..stubs.format_parser import FormatParser

l = logging.getLogger(name=__name__)
l.setLevel('INFO')

class syslog(FormatParser):
    def run(self, priority, fmt):
        fmt = self._parse(fmt)
        formatted = fmt.replace(self.va_arg)
        if not formatted.symbolic:
            formatted = self.state.solver.eval(formatted, cast_to=bytes)
        l.info("Syslog priority %s: %s", priority, formatted)



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_tricore_LE_32_tc176x(ArchPcode):
    name = 'tricore:LE:32:tc176x'
    pcode_arch = 'tricore:LE:32:tc176x'
    description = 'Siemens Tricore Embedded Processor TC1762/TC1766'
    bits = 32
    ip_offset = 0xfe08
    sp_offset = 0xffa8
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('contextreg', 4, 0x0),
        Register('task_asi', 4, 0x8004),
        Register('pma0', 4, 0x8100),
        Register('pma1', 4, 0x8104),
        Register('pma2', 4, 0x8108),
        Register('dcon2', 4, 0x9000),
        Register('dcon1', 4, 0x9008),
        Register('smacon', 4, 0x900c),
        Register('dstr', 4, 0x9010),
        Register('datr', 4, 0x9018),
        Register('deadd', 4, 0x901c),
        Register('diear', 4, 0x9020),
        Register('dietr', 4, 0x9024),
        Register('dcon0', 4, 0x9040),
        Register('pstr', 4, 0x9200),
        Register('pcon1', 4, 0x9204),
        Register('pcon2', 4, 0x9208),
        Register('pcon0', 4, 0x920c),
        Register('piear', 4, 0x9210),
        Register('pietr', 4, 0x9214),
        Register('compat', 4, 0x9400),
        Register('fpu_trap_con', 4, 0xa000),
        Register('fpu_trap_pc', 4, 0xa004),
        Register('fpu_trap_opc', 4, 0xa008),
        Register('fpu_trap_src1', 4, 0xa010),
        Register('fpu_trap_src2', 4, 0xa014),
        Register('fpu_trap_src3', 4, 0xa018),
        Register('dpr0_l', 4, 0xc000),
        Register('dpr0_u', 4, 0xc004),
        Register('dpr1_l', 4, 0xc008),
        Register('dpr1_u', 4, 0xc00c),
        Register('dpr2_l', 4, 0xc010),
        Register('dpr2_u', 4, 0xc014),
        Register('dpr3_l', 4, 0xc018),
        Register('dpr3_u', 4, 0xc01c),
        Register('dpr4_l', 4, 0xc020),
        Register('dpr4_u', 4, 0xc024),
        Register('dpr5_l', 4, 0xc028),
        Register('dpr5_u', 4, 0xc02c),
        Register('dpr6_l', 4, 0xc030),
        Register('dpr6_u', 4, 0xc034),
        Register('dpr7_l', 4, 0xc038),
        Register('dpr7_u', 4, 0xc03c),
        Register('dpr8_l', 4, 0xc040),
        Register('dpr8_u', 4, 0xc044),
        Register('dpr9_l', 4, 0xc048),
        Register('dpr9_u', 4, 0xc04c),
        Register('dpr10_l', 4, 0xc050),
        Register('dpr10_u', 4, 0xc054),
        Register('dpr11_l', 4, 0xc058),
        Register('dpr11_u', 4, 0xc05c),
        Register('dpr12_l', 4, 0xc060),
        Register('dpr12_u', 4, 0xc064),
        Register('dpr13_l', 4, 0xc068),
        Register('dpr13_u', 4, 0xc06c),
        Register('dpr14_l', 4, 0xc070),
        Register('dpr14_u', 4, 0xc074),
        Register('dpr15_l', 4, 0xc078),
        Register('dpr15_u', 4, 0xc07c),
        Register('cpr0_l', 4, 0xd000),
        Register('cpr0_u', 4, 0xd004),
        Register('cpr1_l', 4, 0xd008),
        Register('cpr1_u', 4, 0xd00c),
        Register('cpr2_l', 4, 0xd010),
        Register('cpr2_u', 4, 0xd014),
        Register('cpr3_l', 4, 0xd018),
        Register('cpr3_u', 4, 0xd01c),
        Register('cpr4_l', 4, 0xd020),
        Register('cpr4_u', 4, 0xd024),
        Register('cpr5_l', 4, 0xd028),
        Register('cpr5_u', 4, 0xd02c),
        Register('cpr6_l', 4, 0xd030),
        Register('cpr6_u', 4, 0xd034),
        Register('cpr7_l', 4, 0xd038),
        Register('cpr7_u', 4, 0xd03c),
        Register('cpr8_l', 4, 0xd040),
        Register('cpr8_u', 4, 0xd044),
        Register('cpr9_l', 4, 0xd048),
        Register('cpr9_u', 4, 0xd04c),
        Register('cpr10_l', 4, 0xd050),
        Register('cpr10_u', 4, 0xd054),
        Register('cpr11_l', 4, 0xd058),
        Register('cpr11_u', 4, 0xd05c),
        Register('cpr12_l', 4, 0xd060),
        Register('cpr12_u', 4, 0xd064),
        Register('cpr13_l', 4, 0xd068),
        Register('cpr13_u', 4, 0xd06c),
        Register('cpr14_l', 4, 0xd070),
        Register('cpr14_u', 4, 0xd074),
        Register('cpr15_l', 4, 0xd078),
        Register('cpr15_u', 4, 0xd07c),
        Register('cpxe_0', 4, 0xe000),
        Register('cpxe_1', 4, 0xe004),
        Register('cpxe_2', 4, 0xe008),
        Register('cpxe_3', 4, 0xe00c),
        Register('dpre_0', 4, 0xe010),
        Register('dpre_1', 4, 0xe014),
        Register('dpre_2', 4, 0xe018),
        Register('dpre_3', 4, 0xe01c),
        Register('dpwe_0', 4, 0xe020),
        Register('dpwe_1', 4, 0xe024),
        Register('dpwe_2', 4, 0xe028),
        Register('dpwe_3', 4, 0xe02c),
        Register('tps_con', 4, 0xe400),
        Register('tps_timer0', 4, 0xe404),
        Register('tps_timer1', 4, 0xe408),
        Register('tps_timer2', 4, 0xe40c),
        Register('tr0evt', 4, 0xf000),
        Register('tr0adr', 4, 0xf004),
        Register('tr1evt', 4, 0xf008),
        Register('tr1adr', 4, 0xf00c),
        Register('tr2evt', 4, 0xf010),
        Register('tra2dr', 4, 0xf014),
        Register('tr3evt', 4, 0xf018),
        Register('tr3adr', 4, 0xf01c),
        Register('tr4evt', 4, 0xf020),
        Register('tr4adr', 4, 0xf024),
        Register('tr5evt', 4, 0xf028),
        Register('tr5adr', 4, 0xf02c),
        Register('tr6evt', 4, 0xf030),
        Register('tr6adr', 4, 0xf034),
        Register('tr7evt', 4, 0xf038),
        Register('tr7adr', 4, 0xf03c),
        Register('cctrl', 4, 0xfc00),
        Register('ccnt', 4, 0xfc04),
        Register('icnt', 4, 0xfc08),
        Register('m1cnt', 4, 0xfc0c),
        Register('m2cnt', 4, 0xfc10),
        Register('m3cnt', 4, 0xfc14),
        Register('dbgsr', 4, 0xfd00),
        Register('exevt', 4, 0xfd08),
        Register('crevt', 4, 0xfd0c),
        Register('swevt', 4, 0xfd10),
        Register('trig_acc', 4, 0xfd30),
        Register('dms', 4, 0xfd40),
        Register('dcx', 4, 0xfd44),
        Register('dbgtcr', 4, 0xfd48),
        Register('pcxi', 4, 0xfe00),
        Register('psw', 4, 0xfe04),
        Register('pc', 4, 0xfe08, alias_names=('ip',)),
        Register('syscon', 4, 0xfe14),
        Register('cpu_id', 4, 0xfe18),
        Register('core_id', 4, 0xfe1c),
        Register('biv', 4, 0xfe20),
        Register('btv', 4, 0xfe24),
        Register('isp', 4, 0xfe28),
        Register('icr', 4, 0xfe2c),
        Register('pipn', 1, 0xfe2e),
        Register('fcx', 4, 0xfe38),
        Register('lcx', 4, 0xfe3c),
        Register('e0', 8, 0xff00),
        Register('d0', 4, 0xff00),
        Register('d1', 4, 0xff04),
        Register('e2', 8, 0xff08),
        Register('d2', 4, 0xff08),
        Register('d3', 4, 0xff0c),
        Register('e4', 8, 0xff10),
        Register('d4', 4, 0xff10),
        Register('d5', 4, 0xff14),
        Register('e6', 8, 0xff18),
        Register('d6', 4, 0xff18),
        Register('d7', 4, 0xff1c),
        Register('e8', 8, 0xff20),
        Register('d8', 4, 0xff20),
        Register('d9', 4, 0xff24),
        Register('e10', 8, 0xff28),
        Register('d10', 4, 0xff28),
        Register('d11', 4, 0xff2c),
        Register('e12', 8, 0xff30),
        Register('d12', 4, 0xff30),
        Register('d13', 4, 0xff34),
        Register('e14', 8, 0xff38),
        Register('d14', 4, 0xff38),
        Register('d15', 4, 0xff3c),
        Register('p0', 8, 0xff80),
        Register('a0', 4, 0xff80),
        Register('a1', 4, 0xff84),
        Register('p2', 8, 0xff88),
        Register('a2', 4, 0xff88),
        Register('a3', 4, 0xff8c),
        Register('p4', 8, 0xff90),
        Register('a4', 4, 0xff90),
        Register('a5', 4, 0xff94),
        Register('p6', 8, 0xff98),
        Register('a6', 4, 0xff98),
        Register('a7', 4, 0xff9c),
        Register('p8', 8, 0xffa0),
        Register('a8', 4, 0xffa0),
        Register('a9', 4, 0xffa4),
        Register('p10', 8, 0xffa8),
        Register('a10', 4, 0xffa8),
        Register('a11', 4, 0xffac),
        Register('p12', 8, 0xffb0),
        Register('a12', 4, 0xffb0),
        Register('a13', 4, 0xffb4),
        Register('p14', 8, 0xffb8),
        Register('a14', 4, 0xffb8),
        Register('a15', 4, 0xffbc),
        Register('r0', 4, 0xf0043f00),
        Register('r1', 4, 0xf0043f04),
        Register('r2', 4, 0xf0043f08),
        Register('r3', 4, 0xf0043f0c),
        Register('r4', 4, 0xf0043f10),
        Register('r5', 4, 0xf0043f14),
        Register('r6', 4, 0xf0043f18),
        Register('r7', 4, 0xf0043f1c)
    ]

register_arch(['tricore:le:32:tc176x'], 32, Endness.LE, ArchPcode_tricore_LE_32_tc176x)



import angr

######################################
# open
######################################

class open(angr.SimProcedure): #pylint:disable=W0622
    #pylint:disable=arguments-differ,unused-argument

    def run(self, p_addr, flags, mode):
        strlen = angr.SIM_PROCEDURES['libc']['strlen']

        p_strlen = self.inline_call(strlen, p_addr)
        p_expr = self.state.memory.load(p_addr, p_strlen.max_null_index, endness='Iend_BE')
        path = self.state.solver.eval(p_expr, cast_to=bytes)

        fd = self.state.posix.open(path, flags)
        if fd is None:
            return -1
        return fd



from . import ExplorationTechnique

class LengthLimiter(ExplorationTechnique):
    """
    Length limiter on paths.
    """

    def __init__(self, max_length, drop=False):
        super(LengthLimiter, self).__init__()
        self._max_length = max_length
        self._drop = drop

    def _filter(self, s):
        return s.history.block_count > self._max_length

    def step(self, simgr, stash='active', **kwargs):
        simgr = simgr.step(stash=stash, **kwargs)
        simgr.move('active', '_DROP' if self._drop else 'cut', self._filter)
        return simgr



#!/usr/bin/env python
"""
Creates an archinfo-compatible architecture definition from SLEIGH specs.
"""
# FIXME:
#  - Definition consistency check test cases
#  - Transfering memory space representation information (angr assumes shared code)
#  - Transferring additional details (compiler specs, calling convention, etc)

from archinfo.arch import Register

import os
import os.path
import logging
import pypcode


l = logging.getLogger(__name__)

def find_matching_cid(lang, desired):
    for cid in lang.cspecs:
        if cid[0] == desired:
            return cid
    return None

def create_archinfo_class_from_lang(lang:pypcode.ArchLanguage) -> str:
    """
    Construct an archinfo class definition from pypcode architecture definitions.
    """
    ctx = pypcode.Context(lang)
    archinfo_regs = {
        rname.lower(): Register(rname.lower(), r.size, r.offset)
            for rname, r in ctx.registers.items()
            }

    # Get program counter register
    pc_offset = None
    pc_tag = lang.pspec.find('programcounter')
    if pc_tag is not None:
        pc_reg = pc_tag.attrib.get('register', None)
        if pc_reg is not None:
            # FIXME: Assumes RAM space
            pc_offset = ctx.registers[pc_reg].offset
            aliases = {'pc', 'ip'}
            aliases.discard(pc_reg.lower())
            archinfo_regs[pc_reg.lower()].alias_names = tuple(aliases)

    if pc_offset is None:
        l.warning('Unknown program counter register offset?')
        pc_offset = 0x80000000

    # Get stack pointer register
    sp_offset = None

    if len(lang.cspecs):
        cspec_id = find_matching_cid(lang, 'default') \
                   or find_matching_cid(lang, 'gcc') \
                   or list(lang.cspecs)[0]
        cspec = lang.cspecs[cspec_id]
        sp_tag = cspec.find('stackpointer')
        if sp_tag is not None:
            sp_reg = sp_tag.attrib.get('register', None)
            if sp_reg is not None:
                # FIXME: Assumes RAM space
                sp_offset = ctx.registers[sp_reg].offset

    if sp_offset is None:
        l.warning('Unknown stack pointer register offset?')
        sp_offset = 0x80000008

    bits = int(lang.size)
    archname = lang.id
    endness = {'little': 'Endness.LE', 'big': 'Endness.BE'}[lang.endian]
    def stringify_reg(r):
        return f"Register('{r.name}', {r.size}, {hex(r.vex_offset)}" \
            + ((', alias_names=' + str(r.alias_names)) if len(r.alias_names) else '') \
            + ")"
    reg_list_str = ",\n        ".join([stringify_reg(r) for r in archinfo_regs.values()])
    archname_san = ''.join([c if c.isalnum() else '_' for c in archname])
    classname = 'ArchPcode_' + archname_san
    return (classname, f"""class {classname}(ArchPcode):
    name = '{archname}'
    pcode_arch = '{archname}'
    description = {repr(lang.description)}
    bits = {bits}
    ip_offset = {hex(pc_offset)}
    sp_offset = {hex(sp_offset)}
    bp_offset = sp_offset
    instruction_endness = {endness}
    register_list = [
        {reg_list_str}
    ]

register_arch(['{archname.lower()}'], {bits}, {endness}, {classname})
""")

def main():
    archdir = os.path.dirname(os.path.abspath(__file__))
    langs = [lang for arch in pypcode.Arch.enumerate()
                     for lang in arch.languages]
    imports = []
    for i, lang in enumerate(langs):
        classname, classdef = create_archinfo_class_from_lang(lang)
        imports.append(classname)
        with open(os.path.join(archdir, f'{classname}.py'), 'w') as f:
            print('Generating arch definition for %s (%d of %d)' % (lang.id, i+1, len(langs)))
            f.write(
                "###\n"
                "### This file was automatically generated\n"
                "###\n"
                "\n"
                "from archinfo.arch import register_arch, Endness, Register\n"
                "\n"
                "from .common import ArchPcode\n"
                "\n"
                "\n"
                )
            f.write(classdef)
    with open(os.path.join(archdir, '__init__.py'), 'w') as f:
        for classname in imports:
            f.write(f'from .{classname} import {classname}\n')
        f.write('from .common import *\n')

if __name__ == '__main__':
    main()



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_V850_LE_32_default(ArchPcode):
    name = 'V850:LE:32:default'
    pcode_arch = 'V850:LE:32:default'
    description = 'Renesas V850 family'
    bits = 32
    ip_offset = 0x100
    sp_offset = 0xc
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('r0r1', 8, 0x0),
        Register('r0', 4, 0x0),
        Register('r1', 4, 0x4),
        Register('r2sp', 8, 0x8),
        Register('r2', 4, 0x8),
        Register('sp', 4, 0xc),
        Register('r4r5', 8, 0x10),
        Register('gp', 4, 0x10),
        Register('tp', 4, 0x14),
        Register('r6r7', 8, 0x18),
        Register('r6', 4, 0x18),
        Register('r7', 4, 0x1c),
        Register('r8r9', 8, 0x20),
        Register('r8', 4, 0x20),
        Register('r9', 4, 0x24),
        Register('r10r11', 8, 0x28),
        Register('r10', 4, 0x28),
        Register('r11', 4, 0x2c),
        Register('r12r13', 8, 0x30),
        Register('r12', 4, 0x30),
        Register('r13', 4, 0x34),
        Register('r14r15', 8, 0x38),
        Register('r14', 4, 0x38),
        Register('r15', 4, 0x3c),
        Register('r16r17', 8, 0x40),
        Register('r16', 4, 0x40),
        Register('r17', 4, 0x44),
        Register('r18r19', 8, 0x48),
        Register('r18', 4, 0x48),
        Register('r19', 4, 0x4c),
        Register('r20r21', 8, 0x50),
        Register('r20', 4, 0x50),
        Register('r21', 4, 0x54),
        Register('r22r23', 8, 0x58),
        Register('r22', 4, 0x58),
        Register('r23', 4, 0x5c),
        Register('r24r25', 8, 0x60),
        Register('r24', 4, 0x60),
        Register('r25', 4, 0x64),
        Register('r26r27', 8, 0x68),
        Register('r26', 4, 0x68),
        Register('r27', 4, 0x6c),
        Register('r28r29', 8, 0x70),
        Register('r28', 4, 0x70),
        Register('r29', 4, 0x74),
        Register('ep', 4, 0x78),
        Register('lp', 4, 0x7c),
        Register('eipc', 4, 0x80),
        Register('eipsw', 4, 0x84),
        Register('fepc', 4, 0x88),
        Register('fepsw', 4, 0x8c),
        Register('ecr', 4, 0x90),
        Register('psw', 4, 0x94),
        Register('fpsr', 4, 0x98),
        Register('fpepc', 4, 0x9c),
        Register('fpst', 4, 0xa0),
        Register('fpcc', 4, 0xa4),
        Register('fpcfg', 4, 0xa8),
        Register('sccfg', 4, 0xac),
        Register('scbp', 4, 0xb0),
        Register('eiic', 4, 0xb4),
        Register('feic', 4, 0xb8),
        Register('dbic', 4, 0xbc),
        Register('ctpc', 4, 0xc0),
        Register('ctpsw', 4, 0xc4),
        Register('dbpc', 4, 0xc8),
        Register('dbpsw', 4, 0xcc),
        Register('ctbp', 4, 0xd0),
        Register('dir', 4, 0xd4),
        Register('dbg22', 4, 0xd8),
        Register('dbg23', 4, 0xdc),
        Register('dbg24', 4, 0xe0),
        Register('dbg25', 4, 0xe4),
        Register('dbg26', 4, 0xe8),
        Register('dbg27', 4, 0xec),
        Register('eiwr', 4, 0xf0),
        Register('fewr', 4, 0xf4),
        Register('dbwr', 4, 0xf8),
        Register('bsel', 4, 0xfc),
        Register('pc', 4, 0x100, alias_names=('ip',))
    ]

register_arch(['v850:le:32:default'], 32, Endness.LE, ArchPcode_V850_LE_32_default)



from typing import Optional
import logging

import claripy
import pyvex

from . import irop
from . import ccall
from ..light import VEXMixin
from .... import errors
from .... import sim_options as o

l = logging.getLogger(__name__)
zero = claripy.BVV(0, 32)

def value(ty, val, size: Optional[int]=None):
    if ty == 'Ity_F32':
        return claripy.FPV(float(val), claripy.FSORT_FLOAT)
    elif ty == 'Ity_F64':
        return claripy.FPV(float(val), claripy.FSORT_DOUBLE)
    else:
        if size is not None:
            return claripy.BVV(int(val), size)
        return claripy.BVV(int(val), pyvex.get_type_size(ty))

def symbol(ty, name):
    if ty == 'Ity_F32':
        return claripy.FPS(name, claripy.FSORT_FLOAT)
    elif ty == 'Ity_F64':
        return claripy.FPS(name, claripy.FSORT_DOUBLE)
    else:
        return claripy.BVS(name, pyvex.get_type_size(ty))

class ClaripyDataMixin(VEXMixin):
    """
    This mixin provides methods that makes the vex engine process guest code using claripy ASTs as the data domain.
    """

    # util methods

    def _is_true(self, v):
        return claripy.is_true(v)

    def _is_false(self, v):
        return claripy.is_false(v)

    def _optimize_guarded_addr(self, addr, guard):
        # optimization: is the guard the same as the condition inside the address? if so, unpack the address and remove
        # the guarding condition.
        if isinstance(guard, claripy.ast.Base) and guard.op == 'If' \
                and isinstance(addr, claripy.ast.Base) and addr.op == 'If':
            if guard.args[0] is addr.args[0]:
                # the address is guarded by the same guard! unpack the addr
                return addr.args[1]
        return addr

    # consts

    def _handle_vex_const(self, const):
        return value(const.type, const.value)

    # statements

    def _perform_vex_stmt_LoadG(self, addr, alt, guard, dst, cvt, end):
        addr = self._optimize_guarded_addr(addr, guard)
        super()._perform_vex_stmt_LoadG(addr, alt, guard, dst, cvt, end)

    def _perform_vex_stmt_StoreG(self, addr, data, guard, ty, endness, **kwargs):
        addr = self._optimize_guarded_addr(addr, guard)
        super()._perform_vex_stmt_StoreG(addr, data, guard, ty, endness, **kwargs)

    # is this right? do I care?
    def _handle_vex_expr_GSPTR(self, expr):
        return zero

    def _handle_vex_expr_VECRET(self, expr):
        return zero

    def _handle_vex_expr_Binder(self, expr):
        return zero

    # simple wrappers to implement the fp/bv data casting

    def _perform_vex_expr_Get(self, offset, ty, **kwargs):
        res = super()._perform_vex_expr_Get(offset, ty, **kwargs)
        if ty.startswith('Ity_F'):
            return res.raw_to_fp()
        else:
            return res

    def _perform_vex_expr_Load(self, addr, ty, endness, **kwargs):
        res = super()._perform_vex_expr_Load(addr, ty, endness, **kwargs)
        if ty.startswith('Ity_F'):
            return res.raw_to_fp()
        else:
            return res

    def _perform_vex_stmt_Put(self, offset, data, **kwargs):
        super()._perform_vex_stmt_Put(offset, data.raw_to_bv(), **kwargs)

    def _perform_vex_stmt_Store(self, addr, data, endness, **kwargs):
        super()._perform_vex_stmt_Store(addr, data.raw_to_bv(), endness, **kwargs)

    # op support

    def _perform_vex_expr_ITE(self, cond, ifTrue, ifFalse):
        return claripy.If(cond != 0, ifTrue, ifFalse)

    def _perform_vex_expr_Op(self, op, args):
        # TODO: get rid of these hacks (i.e. state options and modes) and move these switches into engine properties
        options = getattr(self.state, 'options', {o.SUPPORT_FLOATING_POINT})
        simop = irop.vexop_to_simop(
            op,
            extended=o.EXTENDED_IROP_SUPPORT in options,
            fp=o.SUPPORT_FLOATING_POINT in options
        )
        return simop.calculate(*args)

    # ccall support

    def _perform_vex_expr_CCall(self, func_name, ty, args, func=None):
        if func is None:
            try:
                func = getattr(ccall, func_name)
            except AttributeError as e:
                raise errors.UnsupportedCCallError(f"Unsupported ccall {func_name}") from e

        try:
            return func(self.state, *args)
        except ccall.CCallMultivaluedException as e:
            cases, to_replace = e.args
            # pylint: disable=undefined-loop-variable
            for i, arg in enumerate(args):
                if arg is to_replace:
                    break
            else:
                raise errors.UnsupportedCCallError("Trying to concretize a value which is not an argument")
            evaluated_cases = [(case, func(self.state, *args[:i], value_, *args[i+1:])) for case, value_ in cases]
            return claripy.ite_cases(evaluated_cases, value(ty, 0))



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_M8C_BE_16_default(ArchPcode):
    name = 'M8C:BE:16:default'
    pcode_arch = 'M8C:BE:16:default'
    description = 'Cypress M8C Microcontroller Family'
    bits = 16
    ip_offset = 0x10
    sp_offset = 0x2
    bp_offset = sp_offset
    instruction_endness = Endness.BE
    register_list = [
        Register('a', 1, 0x0),
        Register('x', 1, 0x1),
        Register('sp', 1, 0x2),
        Register('f', 1, 0x3),
        Register('pc', 2, 0x10, alias_names=('ip',)),
        Register('contextreg', 4, 0x30)
    ]

register_arch(['m8c:be:16:default'], 16, Endness.BE, ArchPcode_M8C_BE_16_default)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("inkobjcore.dll")
prototypes = \
    {
        # 
        'CreateRecognizer': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pCLSID", "phrec"]),
        # 
        'DestroyRecognizer': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec"]),
        # 
        'GetRecoAttributes': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"dwRecoCapabilityFlags": SimTypeInt(signed=False, label="UInt32"), "awcVendorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 32), "awcFriendlyName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64), "awLanguageId": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 64)}, name="RECO_ATTRS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec", "pRecoAttrs"]),
        # 
        'CreateContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec", "phrc"]),
        # 
        'DestroyContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc"]),
        # 
        'GetResultPropertyList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec", "pPropertyCount", "pPropertyGuid"]),
        # 
        'GetUnicodeRanges': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"wcLow": SimTypeChar(label="Char"), "cChars": SimTypeShort(signed=False, label="UInt16")}, name="CHARACTER_RANGE", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec", "pcRanges", "pcr"]),
        # 
        'AddStroke': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"cbPacketSize": SimTypeInt(signed=False, label="UInt32"), "cPacketProperties": SimTypeInt(signed=False, label="UInt32"), "pPacketProperties": SimTypePointer(SimStruct({"guid": SimTypeBottom(label="Guid"), "PropertyMetrics": SimStruct({"nLogicalMin": SimTypeInt(signed=True, label="Int32"), "nLogicalMax": SimTypeInt(signed=True, label="Int32"), "Units": SimTypeInt(signed=False, label="PROPERTY_UNITS"), "fResolution": SimTypeFloat(size=32)}, name="PROPERTY_METRICS", pack=False, align=None)}, name="PACKET_PROPERTY", pack=False, align=None), offset=0), "cButtons": SimTypeInt(signed=False, label="UInt32"), "pguidButtons": SimTypePointer(SimTypeBottom(label="Guid"), offset=0)}, name="PACKET_DESCRIPTION", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"eM11": SimTypeFloat(size=32), "eM12": SimTypeFloat(size=32), "eM21": SimTypeFloat(size=32), "eM22": SimTypeFloat(size=32), "eDx": SimTypeFloat(size=32), "eDy": SimTypeFloat(size=32)}, name="XFORM", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pPacketDesc", "cbPacket", "pPacket", "pXForm"]),
        # 
        'GetBestResultString': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pcSize", "pwcBestResult"]),
        # 
        'SetGuide': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimStruct({"xOrigin": SimTypeInt(signed=True, label="Int32"), "yOrigin": SimTypeInt(signed=True, label="Int32"), "cxBox": SimTypeInt(signed=True, label="Int32"), "cyBox": SimTypeInt(signed=True, label="Int32"), "cxBase": SimTypeInt(signed=True, label="Int32"), "cyBase": SimTypeInt(signed=True, label="Int32"), "cHorzBox": SimTypeInt(signed=True, label="Int32"), "cVertBox": SimTypeInt(signed=True, label="Int32"), "cyMid": SimTypeInt(signed=True, label="Int32")}, name="RECO_GUIDE", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pGuide", "iIndex"]),
        # 
        'AdviseInkChange': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "bNewStroke"]),
        # 
        'EndInkInput': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc"]),
        # 
        'Process': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pbPartialProcessing"]),
        # 
        'SetFactoid': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "cwcFactoid", "pwcFactoid"]),
        # 
        'SetFlags': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "dwFlags"]),
        # 
        'GetLatticePtr': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypePointer(SimStruct({"ulColumnCount": SimTypeInt(signed=False, label="UInt32"), "pLatticeColumns": SimTypePointer(SimStruct({"key": SimTypeInt(signed=False, label="UInt32"), "cpProp": SimStruct({"cProperties": SimTypeInt(signed=False, label="UInt32"), "apProps": SimTypePointer(SimTypePointer(SimStruct({"guidProperty": SimTypeBottom(label="Guid"), "cbPropertyValue": SimTypeShort(signed=False, label="UInt16"), "pPropertyValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RECO_LATTICE_PROPERTY", pack=False, align=None), offset=0), offset=0)}, name="RECO_LATTICE_PROPERTIES", pack=False, align=None), "cStrokes": SimTypeInt(signed=False, label="UInt32"), "pStrokes": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "cLatticeElements": SimTypeInt(signed=False, label="UInt32"), "pLatticeElements": SimTypePointer(SimStruct({"score": SimTypeInt(signed=True, label="Int32"), "type": SimTypeShort(signed=False, label="UInt16"), "pData": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ulNextColumn": SimTypeInt(signed=False, label="UInt32"), "ulStrokeNumber": SimTypeInt(signed=False, label="UInt32"), "epProp": SimStruct({"cProperties": SimTypeInt(signed=False, label="UInt32"), "apProps": SimTypePointer(SimTypePointer(SimStruct({"guidProperty": SimTypeBottom(label="Guid"), "cbPropertyValue": SimTypeShort(signed=False, label="UInt16"), "pPropertyValue": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="RECO_LATTICE_PROPERTY", pack=False, align=None), offset=0), offset=0)}, name="RECO_LATTICE_PROPERTIES", pack=False, align=None)}, name="RECO_LATTICE_ELEMENT", pack=False, align=None), offset=0)}, name="RECO_LATTICE_COLUMN", pack=False, align=None), offset=0), "ulPropertyCount": SimTypeInt(signed=False, label="UInt32"), "pGuidProperties": SimTypePointer(SimTypeBottom(label="Guid"), offset=0), "ulBestResultColumnCount": SimTypeInt(signed=False, label="UInt32"), "pulBestResultColumns": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), "pulBestResultIndexes": SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)}, name="RECO_LATTICE", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "ppLattice"]),
        # 
        'SetTextContext': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "cwcBefore", "pwcBefore", "cwcAfter", "pwcAfter"]),
        # 
        'SetEnabledUnicodeRanges': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"wcLow": SimTypeChar(label="Char"), "cChars": SimTypeShort(signed=False, label="UInt16")}, name="CHARACTER_RANGE", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "cRanges", "pcr"]),
        # 
        'IsStringSupported': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "wcString", "pwcString"]),
        # 
        'SetWordList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "hwl"]),
        # 
        'GetRightSeparator': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pcSize", "pwcRightSeparator"]),
        # 
        'GetLeftSeparator': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrc", "pcSize", "pwcLeftSeparator"]),
        # 
        'DestroyWordList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwl"]),
        # 
        'AddWordsToWordList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hwl", "pwcWords"]),
        # 
        'MakeWordList': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hrec", "pBuffer", "phwl"]),
        # 
        'GetAllRecognizers': SimTypeFunction([SimTypePointer(SimTypePointer(SimTypeBottom(label="Guid"), offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["recognizerClsids", "count"]),
        # 
        'LoadCachedAttributes': SimTypeFunction([SimTypeBottom(label="Guid"), SimTypePointer(SimStruct({"dwRecoCapabilityFlags": SimTypeInt(signed=False, label="UInt32"), "awcVendorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 32), "awcFriendlyName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64), "awLanguageId": SimTypeFixedSizeArray(SimTypeShort(signed=False, label="UInt16"), 64)}, name="RECO_ATTRS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["clsid", "pRecoAttributes"]),
    }

lib.set_prototypes(prototypes)



from typing import Optional, Dict, Set, Tuple, Generator, TYPE_CHECKING  # pylint:disable=unused-import

import networkx  # pylint:disable=unused-import

import ailment

from ...analysis import Analysis

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function


class MultipleBlocksException(Exception):
    pass


class OptimizationPassStage:
    AFTER_SINGLE_BLOCK_SIMPLIFICATION = 0
    AFTER_GLOBAL_SIMPLIFICATION = 1
    AFTER_VARIABLE_RECOVERY = 2


class OptimizationPass(Analysis):

    ARCHES = [ ]  # strings of supported architectures
    PLATFORMS = [ ]  # strings of supported platforms. Can be one of the following: "win32", "linux"
    STAGE: int = None  # Specifies when this optimization pass should be executed

    def __init__(self, func, blocks_by_addr=None, blocks_by_addr_and_idx=None, graph=None):

        self._func: 'Function' = func
        # self._blocks is just a cache
        self._blocks_by_addr: Dict[Tuple[int,Optional[int]],Set[ailment.Block]] = blocks_by_addr
        self._blocks_by_addr_and_idx: Dict[int,ailment.Block] = blocks_by_addr_and_idx
        self._graph = graph  # type: Optional[networkx.DiGraph]

        # output
        self.out_graph = None  # type: Optional[networkx.DiGraph]

    @property
    def blocks_by_addr(self) -> Dict[Tuple[int,Optional[int]],Set[ailment.Block]]:
        return self._blocks_by_addr

    @property
    def blocks_by_addr_and_idx(self) -> Dict[int,ailment.Block]:
        return self._blocks_by_addr_and_idx

    def analyze(self):

        ret, cache = self._check()
        if ret:
            self._analyze(cache=cache)

    def _check(self):
        """
        Check if this optimization applies to this function.

        :returns: a tuple of (does_apply, cache) where cache is a way to pass
                  information to _analyze so it does not have to be recalculated
        """
        raise NotImplementedError()

    def _analyze(self, cache=None):
        """
        Run the analysis.

        :param cache: information passed from _check so it does not have to be
                      recalculated
        :returns: None
        """
        raise NotImplementedError()

    #
    # Util methods
    #

    def _get_block(self, addr, idx=None) -> Optional[ailment.Block]:

        if not self._blocks_by_addr:
            return None
        else:
            if idx is None:
                blocks = self._blocks_by_addr.get(addr, None)
            else:
                blocks = [ self._blocks_by_addr_and_idx.get((addr, idx), None) ]
            if not blocks:
                return None
            if len(blocks) == 1:
                return next(iter(blocks))
            raise MultipleBlocksException("There are %d blocks at address %#x.%s but only one is requested." % (
                len(blocks), addr, idx
            ))

    def _get_blocks(self, addr, idx=None) -> Generator[ailment.Block,None,None]:
        if not self._blocks_by_addr:
            return
        else:
            if idx is None:
                blocks = self._blocks_by_addr.get(addr, None)
                if blocks is not None:
                    yield from blocks
            else:
                block = self._blocks_by_addr_and_idx.get((addr, idx), None)
                if block is not None:
                    yield block

    def _update_block(self, old_block, new_block):

        if self.out_graph is None:
            self.out_graph = self._graph  # we do not make copy here for performance reasons. we can change it if needed

        if old_block not in self.out_graph:
            return

        in_edges = list(self.out_graph.in_edges(old_block, data=True))
        out_edges = list(self.out_graph.out_edges(old_block, data=True))

        self._remove_block(old_block)
        self.out_graph.add_node(new_block)
        self._blocks_by_addr[new_block.addr].add(new_block)
        self._blocks_by_addr_and_idx[(new_block.addr, new_block.idx)] = new_block

        for src, _, data in in_edges:
            if src is old_block:
                src = new_block
            self.out_graph.add_edge(src, new_block, **data)

        for _, dst, data in out_edges:
            if dst is old_block:
                dst = new_block
            self.out_graph.add_edge(new_block, dst, **data)

    def _remove_block(self, block):

        if self.out_graph is None:
            self.out_graph = self._graph

        if block in self.out_graph:
            self.out_graph.remove_node(block)

        if block.addr in self._blocks_by_addr and block in self._blocks_by_addr[block.addr]:
            self._blocks_by_addr[block.addr].remove(block)
            del self._blocks_by_addr_and_idx[(block.addr, block.idx)]

    @staticmethod
    def _is_add(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Add"

    @staticmethod
    def _is_sub(expr):
        return isinstance(expr, ailment.Expr.BinaryOp) and expr.op == "Sub"



###
### This file was automatically generated
###

from archinfo.arch import register_arch, Endness, Register

from .common import ArchPcode


class ArchPcode_avr8_LE_16_default(ArchPcode):
    name = 'avr8:LE:16:default'
    pcode_arch = 'avr8:LE:16:default'
    description = 'AVR8 with 16-bit word addressable code space'
    bits = 16
    ip_offset = 0x42
    sp_offset = 0x3d
    bp_offset = sp_offset
    instruction_endness = Endness.LE
    register_list = [
        Register('sp', 2, 0x3d),
        Register('spl', 1, 0x3d),
        Register('sph', 1, 0x3e),
        Register('pc', 2, 0x42, alias_names=('ip',)),
        Register('cflg', 1, 0x80),
        Register('zflg', 1, 0x81),
        Register('nflg', 1, 0x82),
        Register('vflg', 1, 0x83),
        Register('sflg', 1, 0x84),
        Register('hflg', 1, 0x85),
        Register('tflg', 1, 0x86),
        Register('iflg', 1, 0x87),
        Register('skip', 1, 0x88),
        Register('contextreg', 4, 0x90),
        Register('r1r0', 2, 0x0),
        Register('r0', 1, 0x0),
        Register('r1', 1, 0x1),
        Register('r3r2', 2, 0x2),
        Register('r2', 1, 0x2),
        Register('r3', 1, 0x3),
        Register('r5r4', 2, 0x4),
        Register('r4', 1, 0x4),
        Register('r5', 1, 0x5),
        Register('r7r6', 2, 0x6),
        Register('r6', 1, 0x6),
        Register('r7', 1, 0x7),
        Register('r9r8', 2, 0x8),
        Register('r8', 1, 0x8),
        Register('r9', 1, 0x9),
        Register('r11r10', 2, 0xa),
        Register('r10', 1, 0xa),
        Register('r11', 1, 0xb),
        Register('r13r12', 2, 0xc),
        Register('r12', 1, 0xc),
        Register('r13', 1, 0xd),
        Register('r15r14', 2, 0xe),
        Register('r14', 1, 0xe),
        Register('r15', 1, 0xf),
        Register('r19r18r17r16', 4, 0x10),
        Register('r17r16', 2, 0x10),
        Register('r16', 1, 0x10),
        Register('r17', 1, 0x11),
        Register('r19r18', 2, 0x12),
        Register('r18', 1, 0x12),
        Register('r19', 1, 0x13),
        Register('r23r22r21r20', 4, 0x14),
        Register('r21r20', 2, 0x14),
        Register('r20', 1, 0x14),
        Register('r21', 1, 0x15),
        Register('r23r22', 2, 0x16),
        Register('r22', 1, 0x16),
        Register('r23', 1, 0x17),
        Register('w', 2, 0x18),
        Register('wlo', 1, 0x18),
        Register('whi', 1, 0x19),
        Register('x', 2, 0x1a),
        Register('xlo', 1, 0x1a),
        Register('xhi', 1, 0x1b),
        Register('y', 2, 0x1c),
        Register('ylo', 1, 0x1c),
        Register('yhi', 1, 0x1d),
        Register('z', 2, 0x1e),
        Register('zlo', 1, 0x1e),
        Register('zhi', 1, 0x1f),
        Register('rampd', 1, 0x58),
        Register('rampx', 1, 0x59),
        Register('rampy', 1, 0x5a),
        Register('rampz', 1, 0x5b),
        Register('sreg', 1, 0x5f)
    ]

register_arch(['avr8:le:16:default'], 16, Endness.LE, ArchPcode_avr8_LE_16_default)



import angr

######################################
# fcntl
######################################

class fcntl(angr.SimProcedure):
    #pylint:disable=arguments-differ

    def run(self, fd, cmd):
        #  this is a stupid stub that does not do anything besides returning an unconstrained variable.
        return self.state.solver.BVS('sys_fcntl', self.arch.sizeof['int'], key=('api', 'fcntl'))



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("avifil32.dll")
prototypes = \
    {
        # 
        'AVIFileInit': SimTypeFunction([], SimTypeBottom(label="Void")),
        # 
        'AVIFileExit': SimTypeFunction([], SimTypeBottom(label="Void")),
        # 
        'AVIFileAddRef': SimTypeFunction([SimTypeBottom(label="IAVIFile")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pfile"]),
        # 
        'AVIFileRelease': SimTypeFunction([SimTypeBottom(label="IAVIFile")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pfile"]),
        # 
        'AVIFileOpenA': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIFile"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppfile", "szFile", "uMode", "lpHandler"]),
        # 
        'AVIFileOpenW': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIFile"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppfile", "szFile", "uMode", "lpHandler"]),
        # 
        'AVIFileInfoW': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypePointer(SimStruct({"dwMaxBytesPerSec": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "dwStreams": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "szFileType": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64)}, name="AVIFILEINFOW", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "pfi", "lSize"]),
        # 
        'AVIFileInfoA': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypePointer(SimStruct({"dwMaxBytesPerSec": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "dwStreams": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwWidth": SimTypeInt(signed=False, label="UInt32"), "dwHeight": SimTypeInt(signed=False, label="UInt32"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "szFileType": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 64)}, name="AVIFILEINFOA", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "pfi", "lSize"]),
        # 
        'AVIFileGetStream': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "ppavi", "fccType", "lParam"]),
        # 
        'AVIFileCreateStreamW': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64)}, name="AVISTREAMINFOW", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "ppavi", "psi"]),
        # 
        'AVIFileCreateStreamA': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 64)}, name="AVISTREAMINFOA", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "ppavi", "psi"]),
        # 
        'AVIFileWriteData': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "ckid", "lpData", "cbData"]),
        # 
        'AVIFileReadData': SimTypeFunction([SimTypeBottom(label="IAVIFile"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile", "ckid", "lpData", "lpcbData"]),
        # 
        'AVIFileEndRecord': SimTypeFunction([SimTypeBottom(label="IAVIFile")], SimTypeInt(signed=True, label="Int32"), arg_names=["pfile"]),
        # 
        'AVIStreamAddRef': SimTypeFunction([SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pavi"]),
        # 
        'AVIStreamRelease': SimTypeFunction([SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=False, label="UInt32"), arg_names=["pavi"]),
        # 
        'AVIStreamInfoW': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64)}, name="AVISTREAMINFOW", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "psi", "lSize"]),
        # 
        'AVIStreamInfoA': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 64)}, name="AVISTREAMINFOA", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "psi", "lSize"]),
        # 
        'AVIStreamFindSample': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lPos", "lFlags"]),
        # 
        'AVIStreamReadFormat': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lPos", "lpFormat", "lpcbFormat"]),
        # 
        'AVIStreamSetFormat': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lPos", "lpFormat", "cbFormat"]),
        # 
        'AVIStreamReadData': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "fcc", "lp", "lpcb"]),
        # 
        'AVIStreamWriteData': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "fcc", "lp", "cb"]),
        # 
        'AVIStreamRead': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lStart", "lSamples", "lpBuffer", "cbBuffer", "plBytes", "plSamples"]),
        # 
        'AVIStreamWrite': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lStart", "lSamples", "lpBuffer", "cbBuffer", "dwFlags", "plSampWritten", "plBytesWritten"]),
        # 
        'AVIStreamStart': SimTypeFunction([SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi"]),
        # 
        'AVIStreamLength': SimTypeFunction([SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi"]),
        # 
        'AVIStreamTimeToSample': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lTime"]),
        # 
        'AVIStreamSampleToTime': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lSample"]),
        # 
        'AVIStreamBeginStreaming': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lStart", "lEnd", "lRate"]),
        # 
        'AVIStreamEndStreaming': SimTypeFunction([SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi"]),
        # 
        'AVIStreamGetFrameOpen': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"biSize": SimTypeInt(signed=False, label="UInt32"), "biWidth": SimTypeInt(signed=True, label="Int32"), "biHeight": SimTypeInt(signed=True, label="Int32"), "biPlanes": SimTypeShort(signed=False, label="UInt16"), "biBitCount": SimTypeShort(signed=False, label="UInt16"), "biCompression": SimTypeInt(signed=False, label="UInt32"), "biSizeImage": SimTypeInt(signed=False, label="UInt32"), "biXPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biYPelsPerMeter": SimTypeInt(signed=True, label="Int32"), "biClrUsed": SimTypeInt(signed=False, label="UInt32"), "biClrImportant": SimTypeInt(signed=False, label="UInt32")}, name="BITMAPINFOHEADER", pack=False, align=None), offset=0)], SimTypeBottom(label="IGetFrame"), arg_names=["pavi", "lpbiWanted"]),
        # 
        'AVIStreamGetFrame': SimTypeFunction([SimTypeBottom(label="IGetFrame"), SimTypeInt(signed=True, label="Int32")], SimTypePointer(SimTypeBottom(label="Void"), offset=0), arg_names=["pg", "lPos"]),
        # 
        'AVIStreamGetFrameClose': SimTypeFunction([SimTypeBottom(label="IGetFrame")], SimTypeInt(signed=True, label="Int32"), arg_names=["pg"]),
        # 
        'AVIStreamOpenFromFileA': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppavi", "szFile", "fccType", "lParam", "mode", "pclsidHandler"]),
        # 
        'AVIStreamOpenFromFileW': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppavi", "szFile", "fccType", "lParam", "mode", "pclsidHandler"]),
        # 
        'AVIStreamCreate': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppavi", "lParam1", "lParam2", "pclsidHandler"]),
        # 
        'AVIMakeCompressedStream': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppsCompressed", "ppsSource", "lpOptions", "pclsidHandler"]),
        # 
        'AVISaveA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["szFile", "pclsidHandler", "lpfnCallback", "nStreams", "pfile", "lpOptions"]),
        # 
        'AVISaveVA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="IAVIStream"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["szFile", "pclsidHandler", "lpfnCallback", "nStreams", "ppavi", "plpOptions"]),
        # 
        'AVISaveW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["szFile", "pclsidHandler", "lpfnCallback", "nStreams", "pfile", "lpOptions"]),
        # 
        'AVISaveVW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeBottom(label="Guid"), offset=0), SimTypePointer(SimTypeFunction([SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["param0"]), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="IAVIStream"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["szFile", "pclsidHandler", "lpfnCallback", "nStreams", "ppavi", "plpOptions"]),
        # 
        'AVISaveOptions': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="IAVIStream"), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0), label="LPArray", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["hwnd", "uiFlags", "nStreams", "ppavi", "plpOptions"]),
        # 
        'AVISaveOptionsFree': SimTypeFunction([SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwKeyFrameEvery": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwBytesPerSecond": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "lpFormat": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbFormat": SimTypeInt(signed=False, label="UInt32"), "lpParms": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "cbParms": SimTypeInt(signed=False, label="UInt32"), "dwInterleaveEvery": SimTypeInt(signed=False, label="UInt32")}, name="AVICOMPRESSOPTIONS", pack=False, align=None), offset=0), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["nStreams", "plpOptions"]),
        # 
        'AVIBuildFilterW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszFilter", "cbFilter", "fSaving"]),
        # 
        'AVIBuildFilterA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpszFilter", "cbFilter", "fSaving"]),
        # 
        'AVIMakeFileFromStreams': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIFile"), offset=0), SimTypeInt(signed=True, label="Int32"), SimTypePointer(SimTypeBottom(label="IAVIStream"), label="LPArray", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["ppfile", "nStreams", "papStreams"]),
        # 
        'AVIMakeStreamFromClipboard': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["cfFormat", "hGlobal", "ppstream"]),
        # 
        'AVIPutFileOnClipboard': SimTypeFunction([SimTypeBottom(label="IAVIFile")], SimTypeInt(signed=True, label="Int32"), arg_names=["pf"]),
        # 
        'AVIGetFromClipboard': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIFile"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lppf"]),
        # 
        'AVIClearClipboard': SimTypeFunction([], SimTypeInt(signed=True, label="Int32")),
        # 
        'CreateEditableStream': SimTypeFunction([SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0), SimTypeBottom(label="IAVIStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["ppsEditable", "psSource"]),
        # 
        'EditStreamCut': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "plStart", "plLength", "ppResult"]),
        # 
        'EditStreamCopy': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "plStart", "plLength", "ppResult"]),
        # 
        'EditStreamPaste': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypePointer(SimTypeInt(signed=True, label="Int32"), offset=0), SimTypeBottom(label="IAVIStream"), SimTypeInt(signed=True, label="Int32"), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "plPos", "plLength", "pstream", "lStart", "lEnd"]),
        # 
        'EditStreamClone': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeBottom(label="IAVIStream"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "ppResult"]),
        # 
        'EditStreamSetNameA': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lpszName"]),
        # 
        'EditStreamSetNameW': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lpszName"]),
        # 
        'EditStreamSetInfoW': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 64)}, name="AVISTREAMINFOW", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lpInfo", "cbInfo"]),
        # 
        'EditStreamSetInfoA': SimTypeFunction([SimTypeBottom(label="IAVIStream"), SimTypePointer(SimStruct({"fccType": SimTypeInt(signed=False, label="UInt32"), "fccHandler": SimTypeInt(signed=False, label="UInt32"), "dwFlags": SimTypeInt(signed=False, label="UInt32"), "dwCaps": SimTypeInt(signed=False, label="UInt32"), "wPriority": SimTypeShort(signed=False, label="UInt16"), "wLanguage": SimTypeShort(signed=False, label="UInt16"), "dwScale": SimTypeInt(signed=False, label="UInt32"), "dwRate": SimTypeInt(signed=False, label="UInt32"), "dwStart": SimTypeInt(signed=False, label="UInt32"), "dwLength": SimTypeInt(signed=False, label="UInt32"), "dwInitialFrames": SimTypeInt(signed=False, label="UInt32"), "dwSuggestedBufferSize": SimTypeInt(signed=False, label="UInt32"), "dwQuality": SimTypeInt(signed=False, label="UInt32"), "dwSampleSize": SimTypeInt(signed=False, label="UInt32"), "rcFrame": SimStruct({"left": SimTypeInt(signed=True, label="Int32"), "top": SimTypeInt(signed=True, label="Int32"), "right": SimTypeInt(signed=True, label="Int32"), "bottom": SimTypeInt(signed=True, label="Int32")}, name="RECT", pack=False, align=None), "dwEditCount": SimTypeInt(signed=False, label="UInt32"), "dwFormatChangeCount": SimTypeInt(signed=False, label="UInt32"), "szName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 64)}, name="AVISTREAMINFOA", pack=False, align=None), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["pavi", "lpInfo", "cbInfo"]),
    }

lib.set_prototypes(prototypes)
